import bpy
from bpy.app.handlers import persistent

previous_selected_objects = []

# Register the handler after the add-on has loaded if the option is enabled
@persistent
def register_handler(dummy=None, depsgraph=None):
    bpy.app.timers.register(lambda: enabled_selection_options(bpy.context.scene, bpy.context))

def selection_option_handler(dummy, depsgraph):
    global previous_selected_objects
    current_selected_objects = bpy.context.selected_objects

    if previous_selected_objects != current_selected_objects:
        previous_selected_objects = current_selected_objects
        if bpy.context.scene.auto_focus_in_outliner or bpy.context.scene.collapse_unselected_collections:
            largest_area = (max((a for a in bpy.context.screen.areas if a.type == "OUTLINER"), key=lambda a:a.width * a.height))
            largest_region = next(r for r in largest_area.regions if r.type == "WINDOW")

            if largest_area and largest_region:
                override_context = {
                    'area': largest_area,
                    'region': largest_region
        }

            # need to check so we for instance is not in preferences or any other area without an outliner
            outliner_areas = [area for area in bpy.context.screen.areas if area.type == "OUTLINER"]
            if not outliner_areas:
                return  # No outliner area found, exit the function

            area = outliner_areas[0]
            override_context['area'] = area
            override_context['region'] = area.regions[-1]

            with bpy.context.temp_override(**override_context):
                if bpy.context.scene.collapse_unselected_collections:
                    bpy.ops.outliner.expanded_toggle()
                    bpy.ops.outliner.expanded_toggle()

                bpy.ops.outliner.show_active()

        def walk_children(ob):
            yield ob
            for child in ob.children:
                yield from walk_children(child)

        if bpy.context.scene.select_children:
            # selects children of selected objects recursively
            for obj in current_selected_objects:
                for child in walk_children(obj):
                    child.select_set(True)

def draw_options_in_outliner(self, context):
    layout = self.layout
    layout.label(text="Selection")
    row = layout.row()
    row.prop(context.scene, 'auto_focus_in_outliner')
    row = layout.row()
    row.prop(context.scene, 'collapse_unselected_collections')
    row = layout.row()
    row.prop(context.scene, 'select_children')

def enabled_selection_options(self, context):
    if self.auto_focus_in_outliner or self.select_children or self.collapse_unselected_collections:
        if selection_option_handler not in bpy.app.handlers.depsgraph_update_post:
            bpy.app.handlers.depsgraph_update_post.append(selection_option_handler)
    else:
        if selection_option_handler in bpy.app.handlers.depsgraph_update_post:
            bpy.app.handlers.depsgraph_update_post.remove(selection_option_handler)

class OutlinerOptions(bpy.types.Operator):
    bl_idname = "keyops.outliner_options"
    bl_label = "KeyOps: Outliner Options"
    bl_description = ""
    bl_options = {'INTERNAL'}

    def register():
        bpy.types.Scene.auto_focus_in_outliner = bpy.props.BoolProperty(
            name="Auto Select in Outliner",
            description="Automatically focus on the selected objects in the outliner, can be slow with many objects",
            default=False,
            update=enabled_selection_options
        )
        bpy.types.Scene.select_children = bpy.props.BoolProperty(
            name="Select Children",
            description="Automatically select children of selected objects",
            default=False,
            update=enabled_selection_options
        )
        bpy.types.Scene.collapse_unselected_collections = bpy.props.BoolProperty(
            name="Collapse Unselected Collections",
            description="Automatically collapse collections that are not related to the selected objects",
            default=False,
            update=enabled_selection_options
        )
        bpy.types.OUTLINER_PT_filter.prepend(draw_options_in_outliner)
        
        bpy.app.timers.register(register_handler)
        bpy.app.handlers.load_post.append(register_handler)

    def unregister():
        if selection_option_handler in bpy.app.handlers.depsgraph_update_post:
            bpy.app.handlers.depsgraph_update_post.remove(selection_option_handler)
        
        handlers = [h.__name__ for h in bpy.app.handlers.load_post]
        if 'register_handler' in handlers:
            bpy.app.handlers.load_post.remove(register_handler)
        
        del bpy.types.Scene.auto_focus_in_outliner
        del bpy.types.Scene.select_children
        del bpy.types.Scene.collapse_unselected_collections

        if bpy.types.OUTLINER_PT_filter:
            bpy.types.OUTLINER_PT_filter.remove(draw_options_in_outliner)
